@@ -116,7 +116,7 @@ class Operators:
             self.operator_list = self.operator_list[:-1]
             if isinstance(arg, dict):
                 for key in arg:
-                    new_name = ".rename(\"{}\")".format(arg[key])
+                    new_name = ".as(\"{}\")".format(arg[key])
                     key_start_index = current_project.index(key)
                     key_end_index = key_start_index + len(key) + 2
                     insert_rename_into_string = current_project[:key_end_index]
@@ -124,6 +124,8 @@ class Operators:
                     insert_rename_into_string += current_project[key_end_index:]
                     current_project = insert_rename_into_string
                 self.operator_list.append(current_project)
+        else:
+            raise WrongQueryError("You can only use rename right after the projection.")
         return self
 
     def zmq_sink(self, host, port):
@@ -159,13 +161,44 @@ class Operators:
         self.operator_list.append(".sink(PrintSinkDescriptor::create())")
         return self
 
+    def file_sink(self, path):
+        """
+        This function adds the a file sink into the operator_list of this class.
+        A file sink prints data in a file found in the given path.
+
+        Returns
+        -------
+        Operators
+
+        """
+        self.operator_list.append(".sink(FileSinkDescriptor::create(\"{}\"))".format(path))
+        return self
+
+    def kafka_sink(self, topic, brokers, timeout):
+        """
+        This function adds the a print sink into the operator_list of this class.
+        A print sink prints data in terminal.
+
+        Returns
+        -------
+        Operators
+
+        """
+        self.operator_list.append(".sink(KAFKASinkDescriptor::create(\"{}\", \"{}\", {}))".format(topic, brokers,
+                                                                                                  timeout))
+        return self
+
+    def compact_window(self, window):
+        self.operator_list.append(window.create_window_query())
+        return self
+
     def window(self, window):
         """
         Adds query to create any window.
 
         Parameters
         ----------
-        window :  Windows
+        window :  WindowsTransformer
                   Window object that manages window type, parameters of window, aggregation function and create query
                   for the window.
 
@@ -282,7 +315,13 @@ class Operators:
         self.operator_list.append(complete_query)
         return self
 
-    def create_query(self, host, port):
+    def window_join(self, window):
+        if "joinWith" in self.operator_list[-1]:
+            if isinstance(window.window, WindowsTransformer):
+                self.operator_list.append(window.create_window_query())
+        return self
+
+    def create_query(self):
         """
         Creates query by joining all operators and finishing the string with the zmq sink.
 
@@ -297,11 +336,14 @@ class Operators:
 
         """
         query = ''.join(self.operator_list)
-        query += ".sink(ZmqSinkDescriptor::create(\"{}\", {}))".format(host, port) + ";"
+        query += ";"
         return query
 
+    def remove_last_operator(self):
+        self.operator_list.pop()
 
-class Windows:
+
+class WindowsTransformer:
     """
     manages the strings we create in DataStream.
     """
@@ -343,9 +385,9 @@ class Windows:
         if len(on) > 0:
             agg_query = ".apply({}(Attribute(\"{}\")))".format(agg_type, on)
         if len(name) != 0:
-            # TODO: uncomment this when NebulaStream supports renaming
-            # agg_query += ".as(Attribute(\"{}\"))".format(name)
-            raise NotSupportedError("NebulaStream does not support renaming columns when aggregating at the moment.")
+            agg_query = agg_query[:-1] # remove last ")"
+            agg_query += "->as(Attribute(\"{}\")))".format(name)
+            # raise NotSupportedError("NebulaStream does not support renaming columns when aggregating at the moment.")
         self.aggregation.append(agg_query)
         return self
 
@@ -407,9 +449,22 @@ class Windows:
         """
         size_units_options = {
             "count": "Count",
-            "sec": "Seconds",
+            "Count": "Count",
             "ms": "Milliseconds",
-            "min": "Minutes"
+            "millis": "Milliseconds",
+            "milliseconds": "Milliseconds",
+            "sec": "Seconds",
+            "s": "Seconds",
+            "second": "Seconds",
+            "seconds": "Seconds",
+            "min": "Minutes",
+            "minute": "Minutes",
+            "minutes": "Minutes",
+            "m": "Minutes",
+            "hours": "Hours",
+            "hour": "Hours",
+            "hr": "Hours",
+            "h": "Hours",
         }
         transformed_event_unit = ""
         if self.event_unit is not None:
@@ -434,6 +489,4 @@ class Windows:
 
         if len(self.aggregation) > 0:
             query += ' '.join(self.aggregation)
-        else:
-            raise InvalidSyntaxError("When applying a window you always need to use an aggregation function too.")
         return query
